%{

/*
   sys.l -- PIKT piktc lexer for the systems.cfg file

   $Header: /var/cvs/PIKT/pikt/src/piktc/sys.l,v 1.53 2005/01/04 13:48:31 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002,2003,2004,2005 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "../lib/common.h"
#include "piktc.h"
#include "sys.tab.h"

char _syslex_[] = "syslex";

#define UPDATE_INCLUDES		((scanpass == 3) && (inctype == INCPROC) && isupdate_includes && infile && ! incdevnull)

#define UPDATE_INCLUDE		if (UPDATE_INCLUDES) fprintf(inctmpout, "%s", yytext);

#define RETURN(token) syslval.sval = mystrdup(yytext, _syslex_); return token

typedef struct {
	int inctyp;
	char *file;
	char *proc;
	char *tmpfile;
	FILE *tmpout;
	FILE *yyin;
	int fd[2];
	YY_BUFFER_STATE bs;
	pid_t pid;
	int indnt;
	int lineno;
} incfil;

incfil sysincfiles[MAXIFS+1];

char *sysincfile = (char *) NULL;
char *sysincproc = (char *) NULL;

char *cp;
int ii, jj;

char *execunput;

char *envar, *enval;

void syserror(const char *);
int syswrap(void);
void beginsysincfile(int, char *, char *);
void pushsysincfile(int, char *, char *, char *, FILE *, FILE *, int *, YY_BUFFER_STATE, pid_t, int, int);
void popsysincfile(int *, char **, char **, char **, FILE **, FILE **, int *, YY_BUFFER_STATE *, pid_t *, int *, int *);

%}

W		[ \t]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]

%%

^{L}[-_\.0-9a-zA-Z]*		{ if (indented || verbatim) {
					yyless(0);
					unput('\177');
					unput('\t');
				  } else {
					UPDATE_INCLUDE;
					RETURN(_STANZA);
				  }
				}

^#{W}*indent{W}*(\177)?\n?	{ lineno++;
				  if (indented) {
					logreturn(PIKTLOG, LOG_WARNING, ERRID_INDENT_WITH_OPEN_INDENT, __FILE__, __LINE__, _syslex_, FALSE, "%s, #indent with earlier open #indent at or near line %d: %s", cfgfilecurr, lineno, yytext);
				  }
				  indented = TRUE;
				}
^#{W}*unindent{W}*(\177)?\n?	{ lineno++;
				  if (! indented) {
					logreturn(PIKTLOG, LOG_WARNING, ERRID_UNINDENT_WITHOUT_OPEN_INDENT, __FILE__, __LINE__, _syslex_, FALSE, "%s, #unindent with no earlier open #indent at or near line %d: %s", cfgfilecurr, lineno, yytext);
				  }
				  indented = FALSE;
				}

^#{W}*(include|verbatim){W}+\<[^\>\]\n]+\>{W}+\[[^\n]+\]{W}*(\177)?\n? {
				  lineno++;
				  verbatim = isverbatim(yytext);
				  *strrchr(yytext, ']') = '\0';
				  cp = strchr(yytext, '[') + 1;
				  sysincproc = mystrdup(cp, _syslex_);
				  *strchr(yytext, '>') = '\0';
				  cp = strchr(yytext, '<') + 1;
				  sysincfile = (char *) mymalloc((strlen(cfgdir)+strlen(cp)+2) * sizeof(char), _syslex_);
				  sprintf(sysincfile, "%s/%s", cfgdir, cp);
				  beginsysincfile(isupdate_includes ? INCPROC : INCFILE, sysincfile, sysincproc);
				}

^#{W}*(include|verbatim){W}+\"[^\"\]\n]+\"{W}+\[[^\n]+\]{W}*(\177)?\n? {
				  lineno++;
				  verbatim = isverbatim(yytext);
				  *strrchr(yytext, ']') = '\0';
				  cp = strchr(yytext, '[') + 1;
				  sysincproc = mystrdup(cp, _syslex_);
				  *cp = '\0';
				  *strrchr(yytext, '\"') = '\0';
				  cp = strchr(yytext, '\"') + 1;
				  sysincfile = mystrdup(cp, _syslex_);
				  beginsysincfile(isupdate_includes ? INCPROC : INCFILE, sysincfile, sysincproc);
				}

^#{W}*(include|verbatim){W}+\[[^\n]+\]{W}*(\177)?\n? {
				  lineno++;
				  verbatim = isverbatim(yytext);
				  *strrchr(yytext, ']') = '\0';
				  cp = strchr(yytext, '[') + 1;
				  sysincproc = mystrdup(cp, _syslex_);
				  sysincfile = (char *) NULL;
				  beginsysincfile(INCPROC, sysincfile, sysincproc);
				}

^#{W}*(include|verbatim){W}+\<[^\>\n]+\>{W}*(\177)?\n? {
				  lineno++;
				  verbatim = isverbatim(yytext);
				  sysincproc = (char *) NULL;
				  *strrchr(yytext, '>') = '\0';
				  cp = strchr(yytext, '<') + 1;
				  sysincfile = (char *) mymalloc((strlen(cfgdir)+strlen(cp)+2) * sizeof(char), _syslex_);
				  sprintf(sysincfile, "%s/%s", cfgdir, cp);
				  beginsysincfile(INCFILE, sysincfile, sysincproc);
				}

^#{W}*(include|verbatim){W}+\"[^\"\n]+\"{W}*(\177)?\n? {
				  lineno++;
				  verbatim = isverbatim(yytext);
				  sysincproc = (char *) NULL;
				  *strrchr(yytext, '\"') = '\0';
				  cp = strchr(yytext, '\"') + 1;
				  sysincfile = mystrdup(cp, _syslex_);
				  beginsysincfile(INCFILE, sysincfile, sysincproc);
				}

^#{W}*echo{W}+\"[^\"\n]+\"{W}*(\177)?\n? {
				  lineno++;
				  *(strrchr(yytext, '\"')+1) = '\0';
				  cp = strchr(yytext, '\"');
				  sysincproc = (char *) mycalloc((size_t)(strlen(cp)+echocmdlen+2), sizeof(char), _syslex_);
				  sprintf(sysincproc, "%s %s", echocmd, cp);
				  sysincfile = (char *) NULL;
				  beginsysincfile(INCPROC, sysincfile, sysincproc);
				}

^#{W}*exec{W}+\[[^\n]+\]{W}*(\177)?\n? {
				  if (scanpass == 3) {
					ii = getexecunput(&execunput, yytext, yyleng, FALSE, FALSE, isexpand_macros) - 1;
					while (ii > -1) {
						unput(*(execunput+ii));
						ii--;
					}
					free(execunput);
				  } else {
					lineno++;
				  }
				}

^#{W}*exec{W}+\"\/dev\/null\"{W}+\[[^\n]+\]{W}*(\177)?\n? {
				  if (scanpass == 3) {
					ii = getexecunput(&execunput, yytext, yyleng, FALSE, TRUE, isexpand_macros) - 1;
					while (ii > -1) {
						unput(*(execunput+ii));
						ii--;
					}
					free(execunput);
				  } else {
					lineno++;
				  }
				}

^#{W}*(piktexec|pexec){W}+\[[^\n]+\]{W}*(\177)?\n? {
				  if (scanpass == 3) {
					ii = getexecunput(&execunput, yytext, yyleng, TRUE, FALSE, isexpand_macros) - 1;
					while (ii > -1) {
						unput(*(execunput+ii));
						ii--;
					}
					free(execunput);
				  } else {
					lineno++;
				  }
				}

^#{W}*(piktexec|pexec){W}+\"\/dev\/null\"{W}+\[[^\n]+\]{W}*(\177)?\n? {
				  if (scanpass == 3) {
					ii = getexecunput(&execunput, yytext, yyleng, TRUE, TRUE, isexpand_macros) - 1;
					while (ii > -1) {
						unput(*(execunput+ii));
						ii--;
					}
					free(execunput);
				  } else {
					lineno++;
				  }
				}

^#{W}*(set|setenv){W}+{L}[_0-9a-zA-Z]*{W}*=?{W}*\"[^\"]*\"{W}*(\177)?\n? {
				  lineno++;
				  *strrchr(yytext, '\"') = '\0';
				  cp = strchr(yytext, '\"') + 1;
				  enval = mystrdup(cp, _syslex_);
				  do {
					cp--;
				  } while (! (isalnum(*cp) || (*cp == '_')));
				  *(cp+1) = '\0';
				  do {
					cp--;
				  } while (! isspace(*cp));
				  envar = mystrdup(cp+1, _syslex_);
				  setenv(envar, enval, TRUE);
				  free(envar);
				  free(enval);
				}

<*>^#{W}*(set|setenv){W}+{L}[_0-9a-zA-Z]*{W}*(\177)?\n? {
				  lineno++;
				  cp = yytext+yyleng;
				  do {
					cp--;
				  } while (! (isalnum(*cp) || (*cp == '_')));
				  *(cp+1) = '\0';
				  do {
					cp--;
				  } while (! isspace(*cp));
				  envar = mystrdup(cp+1, _syslex_);
				  setenv(envar, "1", TRUE);
				  free(envar);
				}

<*>^#{W}*(unset|unsetenv){W}+{L}[_0-9a-zA-Z]*{W}*(\177)?\n? {
				  lineno++;
				  cp = yytext+yyleng;
				  do {
					cp--;
				  } while (! (isalnum(*cp) || (*cp == '_')));
				  *(cp+1) = '\0';
				  do {
					cp--;
				  } while (! isspace(*cp));
				  envar = mystrdup(cp+1, _syslex_);
				  unsetenv(envar);
				  free(envar);
				}

^#.*(\177)?\n			{ lineno++;
				  syserror("syntax error");
				}

{W}*hosts			{ UPDATE_INCLUDE;
				  return _HOSTS;
				}

{W}*members			{ UPDATE_INCLUDE;
				  return _MEMBERS;
				}

{W}*aliases			{ UPDATE_INCLUDE;
				  return _ALIASES;
				}

{L}[-_\.0-9a-zA-Z]*		{ UPDATE_INCLUDE;
				  RETURN(_HOST);
				}

\t\177				{ ; }

{W}				{ UPDATE_INCLUDE; }

(\177)?\n?			{ UPDATE_INCLUDE;
				  lineno++;
				}

.				{ UPDATE_INCLUDE;
				  return yytext[0];
				}

%%

/* sys error handler
 */
void
syserror (const char *message)
{

	if (stzerr)
		return;
	else {
		stzerr = TRUE;
		if (! bailout) {
			if (cfgfileprev && strcmp(cfgfileprev, cfgfilecurr)) {
				logreturn(OUTPUT, LOG_ERR, ERRID_SYS_PARSE_ERROR, __FILE__, __LINE__, "sysparse", FALSE, "%s (or possibly %s), at or near line %d: %s: %s", cfgfilecurr, cfgfileprev, lineno, message, yytext ? yytext : nullstring);
			} else {
				logreturn(OUTPUT, LOG_ERR, ERRID_SYS_PARSE_ERROR, __FILE__, __LINE__, "sysparse", FALSE, "%s, at or near line %d: %s: %s", cfgfilecurr, lineno, message, yytext ? yytext : nullstring);
			}
			fflush(stderr);
		}
		bailout = TRUE;
		while (incnum > 0) {
			endsysincfile();
		}
		return;
	}
}


/* sys end-of-file handler
 */
int
syswrap (void)
{

	if (incnum > 0) {
		if (bailout) {
			while (incnum > 0) {
				endsysincfile();
			}
			return 1;
		} else {
			endsysincfile();
			return 0;
		}
	} else {
		return 1;
	}
}

/* push the current input buffer onto the file stack; create and switch to
   a new input buffer; open the new include file and start feeding its
   contents to the new input buffer; if stripping comments from the new
   include file, spawn a child process to pass the include file
   contents to the strip lexer
 */
void
beginsysincfile (int inctyp, char *file, char *proc)
{
	static char fnnam[] = "beginsysincfile";

	if (scanpass == 1) {
		check_incfile(inctyp, file);
		if (bailout == TRUE)
			return;
	}
	pushsysincfile(inctype, infile, inproc, inctmpfile, inctmpout, sysin, fd, YY_CURRENT_BUFFER, pid, indented, lineno);
	inctype = inctyp;
	infile = file;
	inproc = proc;
	if (file && ! strcmp(file, "/dev/null"))
		incdevnull = TRUE;
	else
		incdevnull = FALSE;
	if (UPDATE_INCLUDES) {
		inctmpfile = (char *) mymalloc((strlen(file)+5) * sizeof(char), fnnam);
		sprintf(inctmpfile, "%s.tmp", file);
		if ((inctmpout = fopen(inctmpfile, "w")) == NULL) {
			logreturn(OUTPUT, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", inctmpfile);
			bailout = TRUE;
			return;
		}
		chmod(inctmpfile, (mode_t)(S_IRUSR|S_IWUSR|S_IRGRP));	/* 640 */
	}
	cfgfileprev = cfgfilecurr;
	if (inctype == INCPROC)
		cfgfilecurr = inproc;
	else
		cfgfilecurr = infile;
	incnum++;
	if (incnum > MAXIFS) {
		logreturn(OUTPUT, LOG_ERR, ERRID_INCLUDE_FILE_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "maximum number (%d) of include files exceeded", MAXIFS);
		bailout = TRUE;
		return;
	}
	/*
	indented = FALSE;
	 */
	lineno = 1;
	if (! verbatim && dostrip) {
		if (pipe(fd) < 0) {
			logreturn(OUTPUT, LOG_ERR, ERRID_PIPE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "pipe() error");
			bailout = TRUE;
			return;
		}
		/* open the ipc channels */
		ipcopen(incnum, __FILE__, __LINE__);
		switch (pid = fork()) {
			case -1:	/* error */
				logreturn(OUTPUT, LOG_ERR, ERRID_FORK_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fork() error");
				bailout = TRUE;
				return;
				break;
			/* pipe from child to parent */
			case 0:		/* child */
				close(fd[0]);
				/* todo: check the return value of instrmopen()
				   and if error, signal parent to bailout
				 */
				stripin = instrmopen(inctype, infile, inproc, incnum, __FILE__, __LINE__);
				stripout = fdopen(fd[1], "w");
				wait_parent(incnum, __FILE__, __LINE__);
				tell_parent(incnum, __FILE__, __LINE__);
				striplex();
				/* todo: check the return value of instrmclose()
				   and if error, signal parent to bailout
				 */
				instrmclose(inctype, stripin, incnum, __FILE__, __LINE__);
				fclose(stripout);
				tell_parent(incnum, __FILE__, __LINE__);
				wait_parent(incnum, __FILE__, __LINE__);
				close(fd[1]);
				exit(0);
				break;
			default:	/* parent */
				close(fd[1]);
				/* the following sleep() is probably not needed
				   but helps prevent hangs on slower, overloaded
				   systems
				sleep(fork_delay);
				 */
				sysin = fdopen(fd[0], "r");
				tell_child(incnum, __FILE__, __LINE__);
				wait_child(incnum, __FILE__, __LINE__);
				break;
		}
	} else {
		pid = 0;
		fd[0] = -1;
		fd[1] = -1;
		if (! (sysin = instrmopen(inctype, infile, inproc, incnum, __FILE__, __LINE__))) {
			bailout = TRUE;
			return;
		}
	}
	yy_switch_to_buffer(yy_create_buffer(sysin, MAXBUF));
	return;
}

/* at the end of the current include file, close the file, delete its
   input buffer, pop the previous include file off the file stack, and
   switch to the previous input buffer
 */
void
endsysincfile (void)
{
	static char fnnam[] = "endsysincfile";
	YY_BUFFER_STATE bs;

	if (! verbatim && dostrip) {
		wait_child(incnum, __FILE__, __LINE__);
		tell_child(incnum, __FILE__, __LINE__);
		close(fd[0]);
		/* close the ipc channels */
		ipcclose(incnum, __FILE__, __LINE__);
	} else {
		if (instrmclose(inctype, sysin, incnum, __FILE__, __LINE__) == ERROR)
			bailout = TRUE;
	}
	if (UPDATE_INCLUDES) {
		if (fclose(inctmpout)) {
			logreturn(OUTPUT, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", inctmpfile);
			bailout = TRUE;
		}
		if (! bailout) {
			if (rename(inctmpfile, infile)) {
				logreturn(OUTPUT, LOG_ERR, ERRID_FILE_RENAME_ERROR, __FILE__, __LINE__, fnnam, TRUE, "rename() error for %s to %s", inctmpfile, infile);
				bailout = TRUE;
			}
		} else {
			unlink(inctmpfile);
		}
	}
	verbatim = FALSE;
	incnum--;
	if (incnum < 0) {
		logreturn(OUTPUT, LOG_ERR, ERRID_INCNUM_LESS_THAN_ZERO, __FILE__, __LINE__, fnnam, FALSE, "incnum < 0");
		bailout = TRUE;
		return;
	}
	yy_delete_buffer(YY_CURRENT_BUFFER);
	popsysincfile(&inctype, &infile, &inproc, &inctmpfile, &inctmpout, &sysin, fd, &bs, &pid, &indented, &lineno);
	cfgfileprev = cfgfilecurr;
	if (inctype == INCPROC)
		cfgfilecurr = inproc;
	else
		cfgfilecurr = infile;
	yy_switch_to_buffer(bs);
	return;
}

/* push current input buffer info onto the file stack
 */
void
pushsysincfile (int inctyp, char *file, char *proc, char *tmpfile, FILE *tmpout, FILE *yyin, int *fd, YY_BUFFER_STATE bs, pid_t pid, int indnt, int lno)
{

	sysincfiles[incnum].inctyp = inctyp;
	sysincfiles[incnum].file = file;
	sysincfiles[incnum].proc = proc;
	sysincfiles[incnum].tmpfile = tmpfile;
	sysincfiles[incnum].tmpout = tmpout;
	sysincfiles[incnum].yyin = yyin;
	sysincfiles[incnum].fd[0] = *fd;
	sysincfiles[incnum].fd[1] = *(fd+1);
	sysincfiles[incnum].bs = bs;
	sysincfiles[incnum].pid = pid;
	sysincfiles[incnum].indnt = indnt;
	sysincfiles[incnum].lineno = lno;
	return;
}

/* pop input buffer info off the file stack
 */
void
popsysincfile (int *inctyp, char **file, char **proc, char **tmpfile, FILE **tmpout, FILE **yyin, int *fdp, YY_BUFFER_STATE *bsp, pid_t *pidp, int *indntp, int *lnp)
{

	*inctyp = sysincfiles[incnum].inctyp;
	*file = sysincfiles[incnum].file;
	*proc = sysincfiles[incnum].proc;
	*tmpfile = sysincfiles[incnum].tmpfile;
	*tmpout = sysincfiles[incnum].tmpout;
	*yyin = sysincfiles[incnum].yyin;
	*fdp = sysincfiles[incnum].fd[0];
	*(fdp+1) = sysincfiles[incnum].fd[1];
	*bsp = sysincfiles[incnum].bs;
	*pidp = sysincfiles[incnum].pid;
	*indntp = sysincfiles[incnum].indnt;
	*lnp = sysincfiles[incnum].lineno;
	return;
}
